import math
from typing import List, Optional

import torch
from stylegan2_torch.equalized_lr import EqualLinear
from stylegan2_torch.generator.conv_block import group_conv, mod
from stylegan2_torch.op.upfirdn2d import upfirdn2d
from stylegan2_torch.utils import make_kernel
from torch import nn
from torch.functional import Tensor
from torch.nn.parameter import Parameter


class Upsample(nn.Module):
    """
    Upsampling + apply blurring FIR filter
    """

    def __init__(self, blur_kernel: List[int], factor: int):
        super().__init__()

        self.factor = factor

        # Factor to compensate for averaging with zeros
        self.kernel: Tensor
        self.register_buffer("kernel", make_kernel(blur_kernel, self.factor))

        # Since upsampling by factor means there is factor - 1 pad1 already built-in
        """
        UPSAMPLE CASE

           kernel: [kkkkk]................[kkkkk] (k_w = 5)
        upsampled:     [x---x---x---x---x---x---] (in_w = 6, up_x = 4)
           padded: [ppppx---x---x---x---x---x---] (pad0 = 4, pad1 = 0)
           output:   [oooooooooooooooooooooooo]   (out_w = 24)
        Hence, pad0 + pad1 = k_w - 1
               pad0 - pad1 = up_x - 1


        DOWNSAMPLE CASE
        
           kernel: [kkkkk]...............[kkkkk] (k_w = 5)
            input:   [xxxxxxxxxxxxxxxxxxxxxxxx]  (in_w = 24)
           padded: [ppxxxxxxxxxxxxxxxxxxxxxxxxp] (pad0 = 2, pad1 = 1)
           output:   [o-o-o-o-o-o-o-o-o-o-o-o]   (out_w = 12)
        Since last (factor - 1) elements are discarded anyway,
        they don't need to be padded
        Hence, pad0 + pad1 = k_w - 1 - (factor - 1)
               pad0 - pad1 = 0 or 1
        """
        p = len(blur_kernel) - factor
        pad0 = (p + 1) // 2 + factor - 1
        pad1 = p // 2

        self.pad = (pad0, pad1)

    def forward(self, input: Tensor) -> Tensor:
        return upfirdn2d(input,
                         self.kernel,
                         up=self.factor,
                         down=1,
                         pad=self.pad)


class ToRGB(nn.Module):

    def __init__(
        self,
        in_channel: int,
        latent_dim: int,
        up: int,
        blur_kernel: List[int],
    ):
        super().__init__()

        # Affine mapping from W to style vector
        self.affine = EqualLinear(latent_dim, in_channel, bias_init=1)

        # Trainable parameters
        self.weight = Parameter(torch.randn(1, 1, in_channel, 1, 1))
        self.scale = 1 / math.sqrt(in_channel)
        self.bias = Parameter(torch.zeros(1, 1, 1, 1))

        if up > 1:
            self.upsample = Upsample(blur_kernel, up)

    def forward(self,
                input: Tensor,
                w: Tensor,
                prev_output: Optional[Tensor] = None) -> Tensor:
        batch, in_channel, _, _ = input.shape

        # Get style vectors (N, 1, C_in, 1, 1)
        style = self.affine(w).view(batch, 1, in_channel, 1, 1)

        # Modulate weights with equalized learning rate (N, C_out, C_in, K_h, K_w)
        weight = mod(self.scale * self.weight, style)

        # Perform convolution and add bias
        out = group_conv(input, weight) + self.bias

        if prev_output is not None:
            out = out + self.upsample(prev_output)

        return out



from contextlib import suppress
from importlib import metadata

from stylegan2_torch.discriminator import Discriminator
from stylegan2_torch.equalized_lr import (Blur, EqualConv2d, EqualLeakyReLU,
                                          EqualLinear)
from stylegan2_torch.generator import Generator
from stylegan2_torch.utils import Resolution, default_channels

__author__ = "Peter Yuen"
__email__ = "ppeetteerrsx@gmail.com"
__version__ = "test"
with suppress(Exception):
    __version__ = metadata.version("stylegan2_torch")


__all__ = [
    "Discriminator",
    "Generator",
    "Resolution",
    "default_channels",
    "Blur",
    "EqualConv2d",
    "EqualLeakyReLU",
    "EqualLinear",
]



import torch
from stylegan2_torch.equalized_lr import EqualLeakyReLU
from torch import nn
from torch.functional import Tensor


class Normalize(nn.Module):
    """
    Normalize latent vector for each sample
    """

    def forward(self, input: Tensor) -> Tensor:
        # input: (N, style_dim)
        # Normalize z in each sample to N(0,1)
        return input * torch.rsqrt(torch.mean(input ** 2, dim=1, keepdim=True) + 1e-8)


class MappingNetwork(nn.Sequential):
    """
    Mapping network from sampling space (z) to disentangled latent space (w)
    """

    def __init__(self, latent_dim: int, n_mlp: int, lr_mlp_mult: float):
        super().__init__(
            Normalize(),
            *[
                EqualLeakyReLU(
                    latent_dim,
                    latent_dim,
                    lr_mult=lr_mlp_mult,
                )
                for _ in range(n_mlp)
            ]
        )



import math
from typing import List, Optional

import torch
from stylegan2_torch.equalized_lr import Blur, EqualLinear
from stylegan2_torch.op.conv2d_gradfix import conv2d, conv_transpose2d
from stylegan2_torch.op.fused_act import FusedLeakyReLU
from torch import nn
from torch.functional import Tensor
from torch.nn.parameter import Parameter


def mod(weight: Tensor, style: Tensor) -> Tensor:
    """
    Modulate convolution weights with style vector
    (styling = scale each input feature map before convolution)

    Args:
        weight (Tensor): (1, C_out, C_in, K_h, K_w)
        style (Tensor): (N, 1, C_in, 1, 1)

    Returns:
        Tensor: (N, C_out, C_in, K_h, K_w)
    """
    return weight * style


def demod(weight: Tensor) -> Tensor:
    """
    Demodulate convolution weights
    (normalization = statistically restore output feature map to unit s.d.)

    Args:
        weight (Tensor): (N, C_out, C_in, K_h, K_w)

    Returns:
        Tensor: (N, C_out, C_in, K_h, K_w)
    """
    batch, out_channel, _, _, _ = weight.shape
    demod = torch.rsqrt(weight.pow(2).sum([2, 3, 4]) + 1e-8).view(
        batch, out_channel, 1, 1, 1
    )
    return weight * demod


def group_conv(input: Tensor, weight: Tensor) -> Tensor:
    """
    Efficiently perform modulated convolution
    (i.e. grouped convolution)

    Args:
        input (Tensor): (N, C_in, H, W)
        weight (Tensor): (N, C_out, C_in, K, K)

    Returns:
        Tensor: (N, C, H + K - 1, W + K - 1)
    """
    batch, in_channel, height, width = input.shape
    _, out_channel, _, k_h, k_w = weight.shape

    weight = weight.view(batch * out_channel, in_channel, k_h, k_w)
    input = input.view(1, batch * in_channel, height, width)

    out = conv2d(input=input, weight=weight, padding=k_h // 2, groups=batch)
    return out.view(batch, out_channel, height, width)


class AddNoise(nn.Module):
    """
    Inject white noise scaled by a learnable scalar (same noise for whole batch)
    """

    def __init__(self):
        super().__init__()

        # Trainable parameters
        self.weight = Parameter(torch.zeros(1))

    def forward(self, input: Tensor, noise: Optional[Tensor]) -> Tensor:
        if noise is None:
            batch, _, height, width = input.shape
            noise = input.new_empty(batch, 1, height, width).normal_()

        return input + self.weight * noise


class ModConvBlock(nn.Module):
    """
    Modulated convolution block

    disentangled latent vector (w) => affine transformation => style vector
    style vector => modulate + demodulate convolution weights => new conv weights
    new conv weights & input features => group convolution => output features
    output features => add noise & leaky ReLU => final output features
    """

    def __init__(
        self, in_channel: int, out_channel: int, kernel_size: int, latent_dim: int
    ):
        super().__init__()

        # Affine mapping from W to style vector
        self.affine = EqualLinear(latent_dim, in_channel, bias_init=1)

        # Trainable parameters
        self.weight = Parameter(
            torch.randn(1, out_channel, in_channel, kernel_size, kernel_size)
        )
        self.scale = 1 / math.sqrt(in_channel * kernel_size ** 2)

        # Noise and Leaky ReLU
        self.add_noise = AddNoise()
        self.leaky_relu = FusedLeakyReLU(out_channel)

    def forward(self, input: Tensor, w: Tensor, noise: Optional[Tensor]) -> Tensor:
        batch, in_channel, _, _ = input.shape

        # Get style vectors (N, 1, C_in, 1, 1)
        style = self.affine(w).view(batch, 1, in_channel, 1, 1)

        # Modulate weights with equalized learning rate (N, C_out, C_in, K_h, K_w)
        weight = mod(self.scale * self.weight, style)

        # Demodulate weights
        weight = demod(weight)

        # Perform convolution
        out = group_conv(input, weight)

        # Add noise
        out = self.add_noise(out, noise)

        # Add learnable bias and activate
        return self.leaky_relu(out)

    def __call__(self, input: Tensor, w: Tensor, noise: Optional[Tensor]) -> Tensor:
        return super().__call__(input, w, noise)


def group_conv_up(input: Tensor, weight: Tensor, up: int = 2) -> Tensor:
    """
    Efficiently perform upsampling + modulated convolution
    (i.e. grouped transpose convolution)

    Args:
        input (Tensor): (N, C_in, H, W)
        weight (Tensor): (N, C_out, C_in, K, K)
        up (int, optional): U. Defaults to 2.

    Returns:
        Tensor: (N, C, (H - 1) * U + K - 1 + 1, (W - 1) * U + K - 1 + 1)
    """
    batch, in_channel, height, width = input.shape
    _, out_channel, _, k_h, k_w = weight.shape

    weight = weight.transpose(1, 2).reshape(batch * in_channel, out_channel, k_h, k_w)
    input = input.view(1, batch * in_channel, height, width)
    out = conv_transpose2d(
        input=input, weight=weight, stride=up, padding=0, groups=batch
    )
    _, _, out_h, out_w = out.shape
    return out.view(batch, out_channel, out_h, out_w)


class UpModConvBlock(nn.Module):
    """
    Modulated convolution block with upsampling

    disentangled latent vector (w) => affine transformation => style vector
    style vector => modulate + demodulate convolution weights => new conv weights
    new conv weights & input features => group convolution and upsampling => output features
    output features => add noise & leaky ReLU => final output features
    """

    def __init__(
        self,
        in_channel: int,
        out_channel: int,
        kernel_size: int,
        latent_dim: int,
        up: int,
        blur_kernel: List[int],
    ):
        super().__init__()

        # Affine mapping from W to style vector
        self.affine = EqualLinear(latent_dim, in_channel, bias_init=1)

        # Trainable parameters
        self.weight = Parameter(
            torch.randn(1, out_channel, in_channel, kernel_size, kernel_size)
        )
        self.scale = 1 / math.sqrt(in_channel * kernel_size ** 2)

        # Blurring kernel
        self.up = up
        self.blur = Blur(blur_kernel, up, kernel_size)

        # Noise and Leaky ReLU
        self.add_noise = AddNoise()
        self.leaky_relu = FusedLeakyReLU(out_channel)

    def forward(self, input: Tensor, w: Tensor, noise: Optional[Tensor]) -> Tensor:
        batch, in_channel, _, _ = input.shape

        # Get style vectors (N, 1, C_in, 1, 1)
        style = self.affine(w).view(batch, 1, in_channel, 1, 1)

        # Modulate weights with equalized learning rate (N, C_out, C_in, K_h, K_w)
        weight = mod(self.scale * self.weight, style)

        # Demodulate weights
        weight = demod(weight)

        # Reshape to use group convolution
        out = group_conv_up(input, weight, self.up)

        # Apply blurring filter for anti-aliasing (linear operation so order doesn't matter?)
        out = self.blur(out)

        # Add noise
        out = self.add_noise(out, noise)

        # Add learnable bias and activate
        return self.leaky_relu(out)

    def __call__(self, input: Tensor, w: Tensor, noise: Optional[Tensor]) -> Tensor:
        return super().__call__(input, w, noise)



import math
from typing import List

from stylegan2_torch.equalized_lr import Blur, EqualConv2d
from stylegan2_torch.op.fused_act import FusedLeakyReLU
from torch import nn
from torch.functional import Tensor


class ConvBlock(nn.Sequential):
    """
    Convolution in feature space

    EqualConv2d: 2D convolution with equalized learning rate
    FusedLeakyReLU: LeakyReLU with a bias added before activation
    """

    def __init__(self, in_channel: int, out_channel: int, kernel_size: int):
        super().__init__(
            EqualConv2d(
                in_channel,
                out_channel,
                kernel_size,
                padding=kernel_size // 2,
                stride=1,
                bias=False,
            ),
            FusedLeakyReLU(out_channel, bias=True),
        )

    def __call__(self, input: Tensor) -> Tensor:
        return super().__call__(input)


class DownConvBlock(nn.Sequential):
    """
    Downsampling convolution in feature space

    Blur: Gaussian filter as low-pass filter for anti-aliasing + adjust tensor shape to preserve downsampled tensor shape
    EqualConv2d: 2D (downsampling) convolution with equalized learning rate
    FusedLeakyReLU: LeakyReLU with a bias added before activation
    """

    def __init__(
        self,
        in_channel: int,
        out_channel: int,
        kernel_size: int,
        down: int,
        blur_kernel: List[int],
    ):
        super().__init__(
            Blur(blur_kernel, -down, kernel_size),
            EqualConv2d(
                in_channel, out_channel, kernel_size, padding=0, stride=down, bias=False
            ),
            FusedLeakyReLU(out_channel, bias=True),
        )

    def __call__(self, input: Tensor) -> Tensor:
        return super().__call__(input)


class RGBDown(nn.Sequential):
    """
    Downsampling convolution in RGB space, hence no need nonlinearity

    Blur: Gaussian filter as low-pass filter for anti-aliasing + adjust tensor shape to preserve downsampled tensor shape
    EqualConv2d: 2D (downsampling) convolution with equalized learning rate
    """

    def __init__(
        self,
        in_channel: int,
        out_channel: int,
        kernel_size: int,
        down: int,
        blur_kernel: List[int],
    ):

        super().__init__(
            Blur(blur_kernel, -down, kernel_size),
            EqualConv2d(
                in_channel, out_channel, kernel_size, padding=0, stride=down, bias=False
            ),
        )

    def __call__(self, input: Tensor) -> Tensor:
        return super().__call__(input)


class ResBlock(nn.Module):
    """
    Residual block

    ConvBlock + DownConvBlock: Convolution + downsampling
    RGBDown: Skip connection from higher (double) resolution RGB image
    """

    def __init__(self, in_channel: int, out_channel: int, blur_kernel: List[int]):
        super().__init__()

        self.conv = ConvBlock(in_channel, in_channel, 3)
        self.down_conv = DownConvBlock(
            in_channel, out_channel, 3, down=2, blur_kernel=blur_kernel
        )
        self.skip = RGBDown(in_channel, out_channel, 1, down=2, blur_kernel=blur_kernel)

    def forward(self, input: Tensor) -> Tensor:
        out = self.conv(input)
        out = self.down_conv(out)
        skip = self.skip(input)
        # sqrt 2 to adhere to equalized learning rate philosophy
        # (i.e. preserve variance in forward pass not initialization)
        return (out + skip) / math.sqrt(2)

    def __call__(self, input: Tensor) -> Tensor:
        return super().__call__(input)



import math
from typing import Any, Dict, List, Literal, Tuple, overload

import torch
from stylegan2_torch.discriminator.blocks import ConvBlock, ResBlock
from stylegan2_torch.equalized_lr import EqualLeakyReLU, EqualLinear
from stylegan2_torch.utils import Resolution, default_channels
from torch import nn
from torch.functional import Tensor


class Discriminator(nn.Module):
    """
    Discriminator module
    """

    def __init__(
        self,
        resolution: Resolution,
        channels: Dict[Resolution, int] = default_channels,
        blur_kernel: List[int] = [1, 3, 3, 1],
    ):
        super().__init__()

        # FromRGB followed by ResBlock
        self.n_layers = int(math.log(resolution, 2))

        self.blocks = nn.Sequential(
            ConvBlock(1, channels[resolution], 1),
            *[
                ResBlock(channels[2 ** i], channels[2 ** (i - 1)], blur_kernel)
                for i in range(self.n_layers, 2, -1)
            ],
        )

        # Minibatch std settings
        self.stddev_group = 4
        self.stddev_feat = 1

        # Final layers
        self.final_conv = ConvBlock(channels[4] + 1, channels[4], 3)
        self.final_relu = EqualLeakyReLU(channels[4] * 4 * 4, channels[4])
        self.final_linear = EqualLinear(channels[4], 1)

    @overload
    def forward(
        self, input: Tensor, *, return_features: Literal[False] = False
    ) -> Tensor:
        ...

    @overload
    def forward(
        self, input: Tensor, *, return_features: Literal[True]
    ) -> Tuple[Tensor, Tensor]:
        ...

    def forward(self, input: Tensor, *, return_features: bool = False):
        # Downsampling blocks
        out: Tensor = self.blocks(input)

        # Minibatch stddev layer in Progressive GAN https://www.youtube.com/watch?v=V1qQXb9KcDY
        # Purpose is to provide variational information to the discriminator to prevent mode collapse
        # Other layers do not cross sample boundaries
        batch, channel, height, width = out.shape
        n_groups = min(batch, self.stddev_group)
        stddev = out.view(
            n_groups, -1, self.stddev_feat, channel // self.stddev_feat, height, width
        )
        stddev = torch.sqrt(stddev.var(0, unbiased=False) + 1e-8)
        stddev = stddev.mean([2, 3, 4], keepdim=True).squeeze(2)
        stddev = stddev.repeat(n_groups, 1, height, width)
        out = torch.cat([out, stddev], 1)

        # Final layers
        out = self.final_conv(out)
        features = self.final_relu(out.view(batch, -1))
        out = self.final_linear(features)

        if return_features:
            return out, features
        else:
            return out

    @overload
    def __call__(
        self, input: Tensor, *, return_features: Literal[False] = False
    ) -> Tensor:
        ...

    @overload
    def __call__(
        self, input: Tensor, *, return_features: Literal[True]
    ) -> Tuple[Tensor, Tensor]:
        ...

    def __call__(self, *args: Any, **kwargs: Any):
        return super().__call__(*args, **kwargs)



import math
import random
from typing import Any, Dict, List, Literal, Optional, Tuple, overload

import torch
from stylegan2_torch.generator.conv_block import ModConvBlock, UpModConvBlock
from stylegan2_torch.generator.mapping import MappingNetwork
from stylegan2_torch.generator.rgb import ToRGB
from stylegan2_torch.utils import Resolution, default_channels
from torch import nn
from torch.functional import Tensor
from torch.nn.parameter import Parameter


class ConstantInput(nn.Module):
    """
    Constant input image
    """

    def __init__(self, channels: int, size: Resolution):
        super().__init__()
        self.input = Parameter(torch.randn(1, channels, size, size))

    def forward(self, input: Tensor) -> Tensor:
        # Broadcast constant input to each sample
        return self.input.repeat(input.shape[0], 1, 1, 1)

    def __call__(self, input: Tensor) -> Tensor:
        return super().__call__(input)


class Generator(nn.Module):
    """
    Generator module
    """

    def __init__(
        self,
        resolution: Resolution,
        latent_dim: int = 512,
        n_mlp: int = 8,
        lr_mlp_mult: float = 0.01,
        channels: Dict[Resolution, int] = default_channels,
        blur_kernel: List[int] = [1, 3, 3, 1],
    ):
        super().__init__()

        self.latent_dim = latent_dim

        # Create mapping network
        self.mapping = MappingNetwork(latent_dim, n_mlp, lr_mlp_mult)

        # Create constant input
        self.input = ConstantInput(channels[4], 4)

        # Create Conv, UpConv and ToRGB Blocks
        self.convs = nn.ModuleList()
        self.up_convs = nn.ModuleList()
        self.to_rgbs = nn.ModuleList()

        self.n_layers = int(math.log(resolution, 2))
        self.n_w_plus = self.n_layers * 2 - 2

        for layer_idx in range(2, self.n_layers + 1):
            # Upsample condition
            upsample = layer_idx > 2

            # Calculate image size and channels at the layer
            prev_layer_size = 2 ** (layer_idx - 1)
            layer_size: Resolution = 2**layer_idx
            layer_channel = channels[layer_size]

            # Upsampling Conv Block
            if upsample:
                self.up_convs.append(
                    UpModConvBlock(
                        channels[prev_layer_size],
                        layer_channel,
                        3,
                        latent_dim,
                        2,
                        blur_kernel,
                    )
                )

            # Normal Conv Block
            self.convs.append(ModConvBlock(layer_channel, layer_channel, 3, latent_dim))

            # ToRGB Block
            self.to_rgbs.append(
                ToRGB(
                    layer_channel,
                    latent_dim,
                    2 if upsample else 1,
                    blur_kernel,
                )
            )

    def mean_latent(self, n_sample: int, device: str) -> Tensor:
        mean_latent = self.mapping(
            torch.randn(n_sample, self.latent_dim, device=device)
        ).mean(0, keepdim=True)
        mean_latent.detach_()
        return mean_latent

    @overload
    def forward(
        self,
        input: List[Tensor],
        *,
        return_latents: Literal[False] = False,
        input_type: Literal["z", "w", "w_plus"] = "z",
        trunc_option: Optional[Tuple[float, Tensor]] = None,
        mix_index: Optional[int] = None,
        noises: Optional[List[Optional[Tensor]]] = None
    ) -> Tensor:
        ...

    @overload
    def forward(
        self,
        input: List[Tensor],
        *,
        return_latents: Literal[True],
        input_type: Literal["z", "w", "w_plus"] = "z",
        trunc_option: Optional[Tuple[float, Tensor]] = None,
        mix_index: Optional[int] = None,
        noises: Optional[List[Optional[Tensor]]] = None
    ) -> Tuple[Tensor, Tensor]:
        ...

    def forward(
        self,
        # Input tensors (N, latent_dim)
        input: List[Tensor],
        *,
        # Return latents
        return_latents: bool = False,
        # Type of input tensor
        input_type: Literal["z", "w", "w_plus"] = "z",
        # Truncation options
        trunc_option: Optional[Tuple[float, Tensor]] = None,
        # Mixing regularization options
        mix_index: Optional[int] = None,
        # Noise vectors
        noises: Optional[List[Optional[Tensor]]] = None
    ):
        # Get w vectors (can have 2 w vectors for mixing regularization)
        ws: List[Tensor]

        if input_type == "z":
            ws = [self.mapping(z) for z in input]
        else:
            ws = input

        # Perform truncation
        if trunc_option:
            trunc_coeff, trunc_tensor = trunc_option
            ws = [trunc_tensor + trunc_coeff * (w - trunc_tensor) for w in ws]

        # Mixing regularization (why add dimension 1 not 0 lol)
        w_plus: Tensor
        if len(ws) == 1:
            # No mixing regularization
            mix_index = self.n_w_plus

            if input_type == "w_plus":
                w_plus = ws[0]
            else:
                w_plus = ws[0].unsqueeze(1).repeat(1, mix_index, 1)

        else:
            mix_index = mix_index if mix_index else random.randint(1, self.n_w_plus - 1)

            w_plus1 = ws[0].unsqueeze(1).repeat(1, mix_index, 1)
            w_plus2 = ws[1].unsqueeze(1).repeat(1, self.n_w_plus - mix_index, 1)

            w_plus = torch.cat([w_plus1, w_plus2], 1)
        # Get noise
        noises_: List[Optional[Tensor]] = (
            noises if noises else [None] * (self.n_w_plus - 1)
        )

        # Constant input
        out = self.input(w_plus)

        # References for this weird indexing:
        # https://github.com/NVlabs/stylegan2-ada-pytorch/issues/50
        # https://github.com/rosinality/stylegan2-pytorch/issues/278
        img = None
        for i in range(self.n_layers - 1):
            if i > 0:
                out = self.up_convs[i - 1](
                    out, w_plus[:, i * 2 - 1], noises_[i * 2 - 1]
                )

            out = self.convs[i](out, w_plus[:, i * 2], noises_[i * 2])
            img = self.to_rgbs[i](out, w_plus[:, i * 2 + 1], img)

        if return_latents:
            return img, w_plus
        else:
            return img

    @overload
    def __call__(
        self,
        input: List[Tensor],
        *,
        return_latents: Literal[False] = False,
        input_type: Literal["z", "w", "w_plus"] = "z",
        trunc_option: Optional[Tuple[float, Tensor]] = None,
        mix_index: Optional[int] = None,
        noises: Optional[List[Optional[Tensor]]] = None
    ) -> Tensor:
        ...

    @overload
    def __call__(
        self,
        input: List[Tensor],
        *,
        return_latents: Literal[True],
        input_type: Literal["z", "w", "w_plus"] = "z",
        trunc_option: Optional[Tuple[float, Tensor]] = None,
        mix_index: Optional[int] = None,
        noises: Optional[List[Optional[Tensor]]] = None
    ) -> Tuple[Tensor, Tensor]:
        ...

    def __call__(self, *args: Any, **kwargs: Any):
        return super().__call__(*args, **kwargs)



import math
from typing import List

import torch
from torch import nn
from torch.functional import Tensor
from torch.nn import functional as F
from torch.nn.parameter import Parameter

from stylegan2_torch.op.conv2d_gradfix import conv2d
from stylegan2_torch.op.fused_act import fused_leaky_relu
from stylegan2_torch.op.upfirdn2d import upfirdn2d
from stylegan2_torch.utils import make_kernel


class EqualConv2d(nn.Module):
    """
    Conv2d with equalized learning rate
    """

    def __init__(
        self,
        in_channel: int,
        out_channel: int,
        kernel_size: int,
        stride: int = 1,
        padding: int = 0,
        bias: bool = True,
    ):
        super().__init__()

        # Equalized Learning Rate
        self.weight = Parameter(
            torch.randn(out_channel, in_channel, kernel_size, kernel_size))
        # std = gain / sqrt(fan_in)
        self.scale = 1 / math.sqrt(in_channel * kernel_size**2)
        self.stride = stride
        self.padding = padding
        self.bias = Parameter(torch.zeros(out_channel)) if bias else None

    def forward(self, input: Tensor) -> Tensor:
        return conv2d(
            input=input,
            weight=self.weight * self.scale,
            bias=self.bias,
            stride=self.stride,
            padding=self.padding,
        )

    def __repr__(self) -> str:
        return (
            f"{self.__class__.__name__}({self.weight.shape[1]}, {self.weight.shape[0]},"
            f" {self.weight.shape[2]}, stride={self.stride}, padding={self.padding})"
        )


class EqualLinear(nn.Module):
    """
    Linear with equalized learning rate
    """

    def __init__(
        self,
        in_dim: int,
        out_dim: int,
        bias_init: int = 0,
        lr_mult: float = 1,
    ):
        super().__init__()

        # Equalized Learning Rate
        self.weight = Parameter(torch.randn(out_dim, in_dim).div_(lr_mult))

        self.bias = Parameter(torch.zeros(out_dim).fill_(bias_init))

        self.scale = (1 / math.sqrt(in_dim)) * lr_mult

        self.lr_mult = lr_mult

    def forward(self, input: Tensor) -> Tensor:
        return F.linear(input,
                        self.weight * self.scale,
                        bias=self.bias * self.lr_mult)

    def __repr__(self) -> str:
        return (
            f"{self.__class__.__name__}({self.weight.shape[1]}, {self.weight.shape[0]})"
        )


class EqualLeakyReLU(nn.Module):
    """
    Leaky ReLU with equalized learning rate
    """

    def __init__(self, in_dim: int, out_dim: int, lr_mult: float = 1):
        super().__init__()

        # Equalized Learning Rate
        self.weight = Parameter(torch.randn(out_dim, in_dim).div_(lr_mult))

        self.bias = Parameter(torch.zeros(out_dim))

        self.scale = (1 / math.sqrt(in_dim)) * lr_mult

        self.lr_mult = lr_mult

    def forward(self, input: Tensor) -> Tensor:
        out = F.linear(input, self.weight * self.scale)
        return fused_leaky_relu(out, self.bias * self.lr_mult)

    def __repr__(self) -> str:
        return (
            f"{self.__class__.__name__}({self.weight.shape[1]}, {self.weight.shape[0]})"
        )


class Blur(nn.Module):

    def __init__(self, blur_kernel: List[int], factor: int, kernel_size: int):
        """
        Apply blurring FIR filter (before / after) a (downsampling / upsampling) op

        Case 1: Upsample (factor > 0)
            Applied after a transpose convolution of stride U and kernel size K

        Args:
            input (Tensor): (N, C, (H - 1) * U + K - 1 + 1, (W - 1) * U + K - 1 + 1)
            blur_kernel (Tensor): FIR filter
            factor (int, optional): U. Defaults to 2.
            kernel_size (int, optional): K. Defaults to 3.

        Returns:
            Tensor: (N, C, H * U, W * U)


        Case 2: Downsample (factor < 0)
            Applied before a convolution of stride U and kernel size K

        Args:
            input (Tensor): (N, C, H, W)
            blur_kernel (Tensor): FIR filter
            factor (int, optional): U. Defaults to 2.
            kernel_size (int, optional): K. Defaults to 3.

        Returns:
            Tensor: (N, C, H - (U + 1) + K - 1, H  - (U + 1) + K - 1)
        """
        super().__init__()

        if factor > 0:
            p = (len(blur_kernel) - factor) - (kernel_size - 1)
            pad0 = (p + 1) // 2 + factor - 1
            pad1 = p // 2 + 1
        else:
            p = (len(blur_kernel) - abs(factor)) + (kernel_size - 1)
            pad0 = (p + 1) // 2
            pad1 = p // 2

        # Factor to compensate for averaging with zeros if upsampling
        self.kernel: Tensor
        self.register_buffer(
            "kernel", make_kernel(blur_kernel, factor if factor > 0 else 1))
        self.pad = (pad0, pad1)

    def forward(self, input: Tensor) -> Tensor:
        return upfirdn2d(input, self.kernel, pad=self.pad)



import math
import random
from typing import Dict, List, Literal, Tuple, Union

import torch
from torch import autograd, nn
from torch.functional import Tensor
from torch.nn import functional as F

from stylegan2_torch.op.conv2d_gradfix import no_weight_gradients

Resolution = Literal[4, 8, 16, 32, 64, 128, 256, 512, 1024, 2048, 4096, 8192]

default_channels: Dict[Resolution, int] = {
    4: 512,
    8: 512,
    16: 512,
    32: 512,
    64: 512,
    128: 256,
    256: 128,
    512: 64,
    1024: 32,
}


def make_kernel(
    k: List[int],
    factor: int = 1,
) -> Tensor:
    """
    Creates 2D kernel from 1D kernel, compensating for upsampling factor
    """
    kernel = torch.tensor(k, dtype=torch.float32)

    kernel = kernel[None, :] * kernel[:, None]

    kernel /= kernel.sum()
    kernel *= factor**2

    return kernel


def accumulate(
    model1: nn.Module,
    model2: nn.Module,
    decay: float = 0.5 ** (32 / (10 * 1000)),
) -> None:
    """
    Accumulate parameters of model2 onto model1 using EMA
    """
    par1 = dict(model1.named_parameters())
    par2 = dict(model2.named_parameters())

    for k in par1.keys():
        par1[k].data.mul_(decay).add_(par2[k].data, alpha=1 - decay)


def d_logistic_loss(
    real_pred: Tensor,
    fake_pred: Tensor,
) -> Tensor:
    # softplus(-f(x)) + softplus(f(x)) is equivalent
    # to original adversarial loss function

    real_loss = F.softplus(-real_pred)
    fake_loss = F.softplus(fake_pred)

    return real_loss.mean() + fake_loss.mean()


def d_r1_loss(real_pred: Tensor, real_img: Tensor) -> Tensor:
    with no_weight_gradients():
        (grad_real,) = autograd.grad(
            outputs=real_pred.sum(), inputs=real_img, create_graph=True
        )
    grad_penalty = grad_real.pow(2).reshape(grad_real.shape[0], -1).sum(1).mean()

    return grad_penalty


def g_nonsaturating_loss(fake_pred: Tensor) -> Tensor:
    loss = F.softplus(-fake_pred).mean()

    return loss


def g_path_regularize(
    fake_img: Tensor,
    latents: Tensor,
    mean_path_length: Union[Tensor, Literal[0]],
    decay: float = 0.01,
) -> Tuple[Tensor, Tensor, Tensor]:
    noise = torch.randn_like(fake_img) / math.sqrt(
        fake_img.shape[2] * fake_img.shape[3]
    )
    (grad,) = autograd.grad(
        outputs=(fake_img * noise).sum(), inputs=latents, create_graph=True
    )
    path_lengths = torch.sqrt(grad.pow(2).sum(2).mean(1))

    path_mean = mean_path_length + decay * (path_lengths.mean() - mean_path_length)

    path_penalty = (path_lengths - path_mean).pow(2).mean()

    return path_penalty, path_mean.detach(), path_lengths


def make_noise(batch: int, latent_dim: int, n_noise: int, device: str):
    if n_noise == 1:
        return torch.randn(batch, latent_dim, device=device)

    return torch.randn(n_noise, batch, latent_dim, device=device).unbind(0)


def mixing_noise(batch: int, latent_dim: int, prob: float, device: str):
    if random.random() < prob:
        return make_noise(batch, latent_dim, 2, device)

    else:
        return [make_noise(batch, latent_dim, 1, device)]


def set_grad_none(model, targets):
    for n, p in model.named_parameters():
        if n in targets:
            p.grad = None



#!/usr/bin/env python

from dunamai import Version
from setuptools import find_packages, setup

setup(
    author="Peter Yuen",
    author_email="ppeetteerrsx@gmail.com",
    python_requires=">=3.8",
    description="A simple, typed, commented Pytorch implementation of StyleGAN2.",
    install_requires=[
        x.strip() for x in open("requirements.txt").readlines() if x.strip() != ""
    ],
    license="MIT license",
    long_description=open("README.md").read(),
    long_description_content_type="text/markdown",
    include_package_data=True,
    keywords="stylegan2-torch",
    name="stylegan2-torch",
    packages=find_packages(
        include=["stylegan2_torch", "stylegan2_torch.*"],
        exclude=["docs"],
    ),
    package_data={
        "stylegan2_torch": ["*.cpp", "*.cu"],
    },
    test_suite="tests",
    url="https://github.com/ppeetteerrs/stylegan2-torch",
    version=Version.from_any_vcs().serialize(),
    zip_safe=False,
    options={"bdist_wheel": {"universal": True}},
)
